from __future__ import absolute_import
from builtins import object
import re
from . import minigraph as mg

class Resolver(object):
    """Class to compute compatible combination of products"""

    def __init__(self, products):
        """Constructor

        :param products: pass the products JSON received from the buyer API call
        """

        # store the products
        self.products = products

        # create a graph (vertices in format product:version and) and create edges with data from compatibility field
        # note that this directed graph by default
        self.graph = mg.MiniGraph()

        if not self.products:
            return

        plugin_tags = dict()
        product_compatible_modules = dict()

        for item in self.products:
            app_type = item["app_type"]
            app_version = item["version"]
            product_node = "{0}:{1}".format(app_type, app_version)
            product_compatible_modules[product_node] = item["compatible_modules"]

            product_tags = item.get('tags', list())
            compatible_module_tags = item.get('compatible_modules_tags', dict())
            for plugin_node in compatible_module_tags.keys():
                for tag in compatible_module_tags[plugin_node]:
                    plugin_node_tag = '{}#{}'.format(plugin_node, tag)
                    # if tag not in product_tags:
                    #     product_tags.append(tag)

                    if plugin_node_tag not in product_tags:
                        product_tags.append(plugin_node_tag)

                    if plugin_node not in plugin_tags:
                        plugin_tags[plugin_node] = dict()
                        plugin_tags[plugin_node]['tags'] = list()
                        plugin_tags[plugin_node]['business_vertical'] = item['business_vertical']

                    plugin_tags[plugin_node]['tags'].append(tag)
                    plugin_tags[plugin_node]['tags'].append('{}#{}'.format(product_node, tag))

            # add the product node and data which we can use later
            self.graph.add_node(product_node, data={'is_plugin': False, 'name': item['name'],
                                                    'business_vertical': item['business_vertical'],
                                                    'tags': product_tags})

        for plugin_node in plugin_tags.keys():
            # add plugin node with data, note that we don't have a friendly name
            # for plugin hence type and name will be same
            plugin_node_parts = plugin_node.split(':')
            name = plugin_node_parts[0] if plugin_node_parts else plugin_node
            tags = list(set(plugin_tags[plugin_node].get('tags', list())))
            business_vertical = plugin_tags[plugin_node]['business_vertical']
            self.graph.add_node(plugin_node, data={'is_plugin': True, 'name': name,
                                                   'business_vertical': business_vertical,
                                                   'tags': tags})

        for product_node, compatible_modules in product_compatible_modules.items():
            # add the edge between product node and plugin node
            for plugin_node in compatible_modules:
                self.graph.add_edge(product_node, plugin_node)

    def get_compatible_combinations(self, query, strict_version_matches=False, include_query_types=False, tags=None):
        """ Method to get bi-directional queries to get compatible combinations between product and plugins

        :param query: pass one or more products or plugins in format [app_type]:[version], [plugin_type]:[version] as list or tuple
        :param strict_version_matches: Check exact version if true or do partial match on version
        :param include_query_types: Include query types in result
        :param tags: Include tags for query
        :return: matches product combinations as a dict ex. "{"hou_redshift": {"versions": ["2.6.37", "2.6.38", "2.6.39"], "is_plugin": true}}
        """

        if not query or not isinstance(query, (str, list, tuple)):
            raise ValueError("query parameter should be list or tuple")

        if isinstance(query, str):
            query = [query]

        if tags and not isinstance(tags, (str, list, tuple)):
            raise ValueError("tags parameter should be list or tuple")

        if tags and isinstance(tags, str):
            tags = [tags]

        if not tags:
            tags = list()

        # get the types from query
        qry_item_types = [item.split(':')[0] for item in query]

        if tags:
            node_tags = self.graph.node(query[0])[1].get('tags', list())
            if not set(tags).issubset(node_tags):
                return dict()

        # # compute matches from the graph edges
        edges = self.graph.edges()

        if strict_version_matches:
            left_vertices_match = [e[1] for e in edges if e[0] in query]
            right_vertices_match = [e[0] for e in edges if e[1] in query]

            # compute matches from the graph edges
            matches = set(left_vertices_match+right_vertices_match)
        else:
            # make regex patterns so we can match partial versions
            re_queries = []
            for item in query:
                # if only the product name has been requested, fix the query to match all versions
                if ":" not in item:
                    item = item+":[0-9]+"
                item += "(\.|$|(?<=[0-9])[v])"
                re_queries.append(re.compile(item)) # check that our pattern either ends in a . or is the end of the string.
                                                                            #   since nuke uses formatting like 10.5v1, we also match v as if it was a .,
                                                                            #   if it is preceded by a number

            # compare each edge to see if it starts with a query
            matches = set()
            for edge in edges:
                for re_query in re_queries:
                    if re_query.match(edge[0]):
                        matches.add(edge[1])
                        matches.add(edge[0])

                    if re_query.match(edge[1]):
                        matches.add(edge[0])
                        matches.add(edge[1])
            # print " > EDGE MATCHES", matches

            # prune the matches to exclude the elements that dont' match
            pruned_matches = set()
            non_queried_matches = set()
            for match in matches:
                qry_item_type = match.split(':')[0]
                for re_query in re_queries:
                    if re_query.match(match):
                        pruned_matches.add(match)

                    elif qry_item_type not in qry_item_types:
                        non_queried_matches.add(match)

                matches = pruned_matches
            # print " > PRUNED MATCHES", matches

            # add back in the nodes we had edges to, but didn't search for explicitly
            for non_queried_match in non_queried_matches:
                for edge in edges:
                    if non_queried_match in edge:
                        if edge[0] in matches or edge[1] in matches:
                            matches.add(non_queried_match)
            # print " > FINAL MATCHES", matches

        # print(" > FINAL MATCHES", matches)

        # filter the matches by tags
        if tags:
            qry_node = self.graph.node(query[0])
            node_tags = qry_node[1].get('tags', list())
            tag_matches = list()
            for match in matches:
                for tag in tags:
                    if '{}#{}'.format(match, tag) in node_tags:
                        tag_matches.append(match)

            other_query_items = query[1:] if len(query) > 1 else list()

            matches = tag_matches \
                if (not other_query_items) or (other_query_items and set(other_query_items).intersection(tag_matches)) \
                else set()

        # return the matches aggregated by version for each product type
        result = dict()

        for item in matches:
            item_type, item_version = item.split(':')

            # matches list includes the items from query
            # don't need to include any item with the same product type as in query
            if item_type not in qry_item_types or include_query_types:
                # get node data
                node_data = self.graph.node(item)[1]

                if item_type not in result:
                    result[item_type] = dict()
                    result[item_type]['is_plugin'] = node_data.get('is_plugin', False)
                    result[item_type]['name'] = node_data.get('name')
                    result[item_type]['business_vertical'] = node_data.get('business_vertical')
                    result[item_type]['versions'] = list()

                result[item_type]['versions'].append(item_version)

        # sorting of versions done using distutils.version
        # sort the versions in each
        for item_type, version_items in result.items():
            result[item_type]['versions'] = sorted(version_items['versions'], key=ver_cmp_key())
        
        return result

    def get_versions_by_type(self, query, tags=None):
        """ method to get the list of version for a item type

        :param query: query is either an app_type for a product or plugin_type for a plugin
        :param tags: Include tags for query
        :return: returns an array of version for an item type
        """

        result = []

        edges = self.graph.edges()

        # note it is an O(n) iteration for now, we can re-look at this later if we are dealing with a very large list
        for item in self.graph.nodes():
            item_type, item_version = item[0].split(':')

            if item_type == query:
                result.append(item_version)

        if tags:
            tag_result = list()
            for item in result:
                node = '{}:{}'.format(query, item)
                node_tags = self.graph.node(node)[1].get('tags', list())
                for tag in tags:
                    if tag in node_tags:
                        tag_result.append(item)
                        break

            result = tag_result

        # sorting of versions done using distutils.version
        return sorted(result, key=ver_cmp_key())

    def get_all_types(self, exclude_plugins=False):
        """ get all the types and their respective versions
        "return: returns a dict { "hou": { "versions": ["17.5.173", "17.5.229"], "is_plugin": false } }
        """

        result = {}

        for item in self.graph.nodes():
            # get the node data
            node_data = item[1]

            if node_data.get('is_tag'):
                continue

            item_type, item_version = item[0].split(':')

            if exclude_plugins and node_data.get('is_plugin'):
                continue

            if item_type not in result:
                print(node_data)
                result[item_type] = dict()
                result[item_type]['is_plugin'] = node_data.get('is_plugin', False)
                result[item_type]['name'] = node_data.get('name')
                result[item_type]['business_vertical'] = node_data.get('business_vertical')
                result[item_type]['versions'] = list()
            
            result[item_type]['versions'].append(item_version)

        # sorting of versions done using distutils.version
        for item_type, version_items in result.items():
            result[item_type]['versions'] = sorted(version_items['versions'], key=ver_cmp_key())

        return result


def ver_cmp_key():
    'Used as key function for version sorting'

    class K:
        def __init__(self, obj):
            self.obj = obj

        def __lt__(self, other):
            return self.__compare(other) < 0

        def __gt__(self, other):
            return self.__compare(other) > 0

        def __eq__(self, other):
            return self.__compare(other) == 0

        @staticmethod
        def __is_numeric(s):
            return all(c.isdigit() or c == '.' for c in s)

        def __compare(self, other):
            xitems = self.obj.split(".")
            if "-" in xitems[-1]:
                xitems = xitems[:-1] + xitems[-1].split("-")
            yitems = other.obj.split(".")
            if "-" in yitems[-1]:
                yitems = yitems[:-1] + yitems[-1].split("-")
            big = max(len(xitems), len(yitems))
            for i in range(big):
                xitem, yitem = None, None
                try:
                    xitem = xitems[i]
                except IndexError:
                    return -1
                try:
                    yitem = yitems[i]
                except IndexError:
                    return 1
                if self.__is_numeric(xitem) and self.__is_numeric(yitem):
                    xitem = int(xitem)
                    yitem = int(yitem)
                elif self.__is_numeric(xitem) and not self.__is_numeric(yitem):
                    return 1
                elif not self.__is_numeric(xitem) and self.__is_numeric(yitem):
                    return -1
                if xitem > yitem:
                    return 1
                elif xitem < yitem:
                    return -1
                else:
                    continue
            return 0

    return K
